module Typecheck where -- (typeCheckExp, typeCheckStm, typeCheckVar, State) where 


import Abssyntax
import Control.Monad.State hiding (State)
import Data.Map as Map
import Prelude hiding (lookup)

type Types = [Map Ident Type]
type Function = (Type, [Type])

data State = State {variables::Types,functions::[Map Ident Function],function::Ident}


inList :: Eq a => a -> [a] -> Bool
inList _ [] = False
inList a (x:xs) = if a == x then True else inList a xs

assert :: Monad m => Bool -> String -> m ()
assert True _ = return ()
assert False s = fail s

typeCheckExp :: (MonadState State m) => Exp -> m Type
typeCheckExp (BiOpExp e o e') = do
	t1 <- typeCheckExp e
	t2 <- typeCheckExp e'
	assert (t1 == t2) "The parameters for the binary operator aren't equal"
	if inList o [Eq,NEq] then return TBool
		else do 
			assert (t1 == TInt) "The parameters need to be of type int" 
			if inList o [Plus,Minus,Times,Div]
				then return TInt
				else return TBool
typeCheckExp (EVar i) = typeCheckVar i
typeCheckExp (EAss i e) = do
	t <- typeCheckVar i
	t2 <- typeCheckExp e
	assert (t == t2) $ "Illegal to assign an expression of type "++show t2++" to variable "++show i++" of type "++show t 
	return t
typeCheckExp (EInt i) = return TInt
typeCheckExp (EBool b) = return TBool
typeCheckExp EReadI = return TInt
typeCheckExp EReadB = return TBool
typeCheckExp (EPost i op) = do
	TInt <- typeCheckVar i
	return TInt
typeCheckExp (ENeg e) = do
	TInt <- typeCheckExp e
	return TInt
typeCheckExp (ENot e) = do
	TBool <- typeCheckExp e
	return TBool

typeCheckVar :: (MonadState State m) => Ident -> m Type 
typeCheckVar i = do
	s <- get
	findVariable i $ variables s

findVariable :: (MonadState State m) => Ident -> Types -> m Type
findVariable i [] = fail $ "Variable "++show i++" not found in any scope."
findVariable i (m:ms) = if member i m then lookup i m else findVariable i ms

pushAndPop :: (MonadState State m) => m a -> m ()
pushAndPop s = do
	modify (\s -> s { variables = empty:variables s})
	s
	modify (\s -> s { variables = tail $ variables s})

typeCheckStm :: (MonadState State m) => Stm -> m ()
typeCheckStm SNoop = return ()
typeCheckStm (SExp e) = do 
	typeCheckExp e
	return ()
typeCheckStm (SBlock ss) = pushAndPop $ mapM typeCheckStm ss
typeCheckStm (SIf e s s') = do
	TBool <- typeCheckExp e
	pushAndPop $ typeCheckStm s
	pushAndPop $ typeCheckStm s'
typeCheckStm (SWhile e s) = do
	TBool <- typeCheckExp e
	pushAndPop $ typeCheckStm s
typeCheckStm (SDeclD t i) = addVariable i t
typeCheckStm (SDecl t i e) = do
	t2 <- typeCheckExp e
	assert (t == t2) $ "Illegal to assign an expression of type "++show t2++" to variable "++show i++" of type "++show t
	addVariable i t
typeCheckStm (SPrint e) = do
	typeCheckExp e
	return ()

addVariable :: (MonadState State m) => Ident -> Type -> m ()
addVariable i t = do
	s <- get
	let (m:ms) = variables s in case insertLookupWithKey (\k a1 a2 -> a1) i t m of
		(Nothing,m') -> modify (\s -> s{ variables = m':ms})
		_ -> fail $ "Duplicate variable declaration: "++show i
